import json
from typing import Literal
from pathlib import Path

import matplotlib.pyplot as plt
import pandas as pd

from .draw import Chart as BaseChart
from .cursor import Chart as BaseCursorChart, format_candleinfo_ko, format_volumeinfo_ko, format_candleinfo_en, format_volumeinfo_en
from .slider import Chart as BaseSliderChart


__all__ = [
    'path_samsung', 'path_apple',
    'format_candleinfo_ko', 'format_volumeinfo_ko',
    'format_candleinfo_en', 'format_volumeinfo_en',
    'sample', 'switch_backend', 'show', 'close',
    'Chart', 'CursorChart', 'SliderChart',
]


path_samsung = Path(__file__).parent / 'sample/samsung.txt'
path_apple = Path(__file__).parent / 'sample/apple.txt'

def sample(stock: Literal['samsung', 'apple']='samsung', chart: Literal['Chart', 'CursorChart', 'SliderChart']='SliderChart'):
    C: BaseSliderChart = {'Chart': BaseChart, 'CursorChart': BaseCursorChart, 'SliderChart': BaseSliderChart}[chart]()
    path_file = path_samsung if stock == 'samsung' else path_apple
    if stock == 'samsung':
        C.format_candleinfo = format_candleinfo_ko
        C.format_volumeinfo = format_volumeinfo_ko
    else:
        C.format_candleinfo = format_candleinfo_en
        C.format_volumeinfo = format_volumeinfo_en
        C.unit_price = '$'
        C.unit_volume = 'Vol'
        C.digit_price = 3
        C.format_ma = 'ma{}'

    with open(path_file, 'r', encoding='utf-8') as txt:
        data = json.load(txt)
    df = pd.DataFrame(data)

    C.set_data(df)

    show()
    close()
    return


def switch_backend(newbackend='TkAgg'):
    "call matplotlib.pyplot.switch_backend(newbackend)"
    return plt.switch_backend(newbackend)


def show(Close=False):
    """
    call matplotlib.pyplot.show()
    ```if Close``` if True, run matplotlib.pyplot.close('all') after window closee.
    """
    plt.show()
    if Close: close()
    return


def close(fig='all'):
    "call matplotlib.pyplot.close(fig)"
    return plt.close(fig)


class Chart(BaseChart):
    r"""
    You can see the guidance document:
        Korean: https://white.seolpyo.com/entry/147/
        English: https://white.seolpyo.com/entry/148/

    Quick Start:
        ```
        import seolpyo_mplchart as mc
        chart = mc.SliderChart() # Create instance
        chart.set_data(df) # set stock price data
        mc.show() # show chart(run ```matplotlib.pyplot.show()```)
        mc.close() # run ```matplotlib.pyplot.close('close')```
        ```

    Class Variables:
        watermark: watermark text.

        figsize: Default size when creating a matplotlib window
        ratio_ax_legend, ratio_ax_price, ratio_ax_volume: Axes ratio
        adjust: figure adjust. default ```dict(top=0.95, bottom=0.05, left=0.01, right=0.93, wspace=0, hspace=0)```.
        color_background: color of background. default '#fafafa'.
        gridKwargs: kwargs applied to the grid
        color_tick, color_tick_label: Tick and tick label colors. default ('k', 'k').

        df: stock data DataFrame.
        date: date column key. default 'date'
        Open, high, low, close: price column key. default ('open', 'high', 'low', 'close')
        volume: volume column key. default 'volume'. If ```if self.volume``` is ```False```, the volume chart is not drawn.

        format_ma: moving average legend label format. default '{}일선'
        list_ma: Decide how many days to draw the moving average line. default (5, 20, 60, 120, 240)
        list_macolor: Color the moving average line. If the number of colors is greater than the moving average line, black is applied

        candle_on_ma: Decide whether to draw candles on the moving average line. default ```True```

        color_up: The color of the candle. When the closing price is greater than the opening price. default '#fe3032'
        color_down: The color of the candle. When the opening price is greater than the opening price. default '#0095ff'
        color_flat: The color of the candle. WWhen the closing price is the same as the opening price. default 'k'
        color_up_down: The color of the candle. If the closing price is greater than the opening price, but is lower than the previous day's closing price. default 'w'
        color_down_up: The color of the candle. If the opening price is greater than the closing price, but is higher than the closing price of the previous day. default 'w'
        colors_volume: The color of the volume bar. default '#1f77b4'

        color_priceline: The color of the price line. default 'k'

        facecolor_volume, edgecolor_volume = The color of the volume. default ('#1f77b4', 'k')

        candle_width_half, volume_width_half: half of the thickness of the candle and volume

        color_box: The color of the candlebox and volumebox. Used when the mouse is over a candle or volume bar. default 'k'

        limit_candle: Maximum number of candles to draw. default 800
        limit_wick:  Maximum number of candle wicks to draw. default 4,000
    """
    pass


class CursorChart(BaseCursorChart):
    r"""
    You can see the guidance document:
        Korean: https://white.seolpyo.com/entry/147/
        English: https://white.seolpyo.com/entry/148/

    Quick Start:
        ```
        import seolpyo_mplchart as mc
        chart = mc.SliderChart() # Create instance
        chart.set_data(df) # set stock price data
        mc.show() # show chart(run ```matplotlib.pyplot.show()```)
        mc.close() # run ```matplotlib.pyplot.close('close')```
        ```

    Class Variables:
        watermark: watermark text.

        unit_price, unit_volume: price and volume unit. default ('원', '주').
        digit_price, digit_volume: display decimal places when displaying price and volume. default (0, 0),

        figsize: Default size when creating a matplotlib window
        ratio_ax_legend, ratio_ax_price, ratio_ax_volume: Axes ratio
        adjust: figure adjust. default ```dict(top=0.95, bottom=0.05, left=0.01, right=0.93, wspace=0, hspace=0)```.
        color_background: color of background. default '#fafafa'.
        gridKwargs: kwargs applied to the grid
        color_tick, color_tick_label: Tick and tick label colors. default ('k', 'k').

        df: stock data DataFrame.
        date: date column key. default 'date'
        Open, high, low, close: price column key. default ('open', 'high', 'low', 'close')
        volume: volume column key. default 'volume'. If ```if self.volume``` is ```False```, the volume chart is not drawn.

        format_ma: moving average legend label format. default '{}일선'
        list_ma: Decide how many days to draw the moving average line. default (5, 20, 60, 120, 240)
        list_macolor: Color the moving average line. If the number of colors is greater than the moving average line, black is applied

        candle_on_ma: Decide whether to draw candles on the moving average line. default ```True```

        color_up: The color of the candle. When the closing price is greater than the opening price. default '#fe3032'
        color_down: The color of the candle. When the opening price is greater than the opening price. default '#0095ff'
        color_flat: The color of the candle. WWhen the closing price is the same as the opening price. default 'k'
        color_up_down: The color of the candle. If the closing price is greater than the opening price, but is lower than the previous day's closing price. default 'w'
        color_down_up: The color of the candle. If the opening price is greater than the closing price, but is higher than the closing price of the previous day. default 'w'
        colors_volume: The color of the volume bar. default '#1f77b4'

        color_priceline: The color of the price line. default 'k'

        facecolor_volume, edgecolor_volume = The color of the volume. default ('#1f77b4', 'k')

        candle_width_half, volume_width_half: half of the thickness of the candle and volume

        color_box: The color of the candlebox and volumebox. Used when the mouse is over a candle or volume bar. default 'k'

        lineKwargs: kwarg applied to lines drawn based on mouse cursor position
        textboxKwargs: kwarg applied to the info text bbox drawn on the chart. When this is applied, the following occurs: ```textKwargs['bbox'] = textboxKwargs```
        textKwargs: A kwarg that applies to the informational text drawn on the chart. When this is applied, the following occurs: ```textKwargs['bbox'] = textboxKwargs```

        fraction: Decide whether to express information as a fraction. default False
        format_candleinfo: Candle information text format. default '{dt}\n\n종가:　 {close}\n등락률: {rate}\n대비:　 {compare}\n시가:　 {open}({rate_open})\n고가:　 {high}({rate_high})\n저가:　 {low}({rate_low})\n거래량: {volume}({rate_volume})'
        format_volumeinfo: Volume information text format. default '{dt}\n\n거래량　　　: {volume}\n거래량증가율: {rate_volume}'

        limit_candle: Maximum number of candles to draw. default 800
        limit_wick:  Maximum number of candle wicks to draw. default 4,000
    """
    pass


class SliderChart(BaseSliderChart):
    r"""
    You can see the guidance document:
        Korean: https://white.seolpyo.com/entry/147/
        English: https://white.seolpyo.com/entry/148/

    Quick Start:
        ```
        import seolpyo_mplchart as mc
        chart = mc.SliderChart() # Create instance
        chart.set_data(df) # set stock price data
        mc.show() # show chart(run ```matplotlib.pyplot.show()```)
        mc.close() # run ```matplotlib.pyplot.close('close')```
        ```

    Class Variables:
        watermark: watermark text.

        unit_price, unit_volume: price and volume unit. default ('원', '주').
        digit_price, digit_volume: display decimal places when displaying price and volume. default (0, 0),

        figsize: Default size when creating a matplotlib window
        slider_top: ax_slider is located at the top or bottom. default ```True```.
        ratio_ax_slider, ratio_ax_legend, ratio_ax_price, ratio_ax_volume: Axes ratio
        ratio_ax_none: Ratio between volume chart and slider. Used only when slider_top is ```False```
        adjust: figure adjust. default ```dict(top=0.95, bottom=0.05, left=0.01, right=0.93, wspace=0, hspace=0)```.
        color_background: color of background. default '#fafafa'.
        gridKwargs: kwargs applied to the grid
        color_tick, color_tick_label: Tick and tick label colors. default ('k', 'k').

        df: stock data DataFrame.
        date: date column key. default 'date'
        Open, high, low, close: price column key. default ('open', 'high', 'low', 'close')
        volume: volume column key. default 'volume'. If ```if self.volume``` is ```False```, the volume chart is not drawn.

        format_ma: moving average legend label format. default '{}일선'
        list_ma: Decide how many days to draw the moving average line. default (5, 20, 60, 120, 240)
        list_macolor: Color the moving average line. If the number of colors is greater than the moving average line, black is applied

        candle_on_ma: Decide whether to draw candles on the moving average line. default ```True```
        color_sliderline: Color of closing price line in ax_slider. default 'k'
        color_navigatorline: Color of left and right dividing lines in selected area. default '#1e78ff'
        color_navigator: Color of unselected area. default 'k'

        color_up: The color of the candle. When the closing price is greater than the opening price. default '#fe3032'
        color_down: The color of the candle. When the opening price is greater than the opening price. default '#0095ff'
        color_flat: The color of the candle. WWhen the closing price is the same as the opening price. default 'k'
        color_up_down: The color of the candle. If the closing price is greater than the opening price, but is lower than the previous day's closing price. default 'w'
        color_down_up: The color of the candle. If the opening price is greater than the closing price, but is higher than the closing price of the previous day. default 'w'
        colors_volume: The color of the volume bar. default '#1f77b4'

        color_priceline: The color of the price line. default 'k'

        facecolor_volume, edgecolor_volume = The color of the volume. default ('#1f77b4', 'k')

        candle_width_half, volume_width_half: half of the thickness of the candle and volume

        color_box: The color of the candlebox and volumebox. Used when the mouse is over a candle or volume bar. default 'k'

        lineKwargs: kwarg applied to lines drawn based on mouse cursor position
        textboxKwargs: kwarg applied to the info text bbox drawn on the chart. When this is applied, the following occurs: ```textKwargs['bbox'] = textboxKwargs```
        textKwargs: A kwarg that applies to the informational text drawn on the chart. When this is applied, the following occurs: ```textKwargs['bbox'] = textboxKwargs```

        fraction: Decide whether to express information as a fraction. default False
        format_candleinfo: Candle information text format. default '{dt}\n\n종가:　 {close}\n등락률: {rate}\n대비:　 {compare}\n시가:　 {open}({rate_open})\n고가:　 {high}({rate_high})\n저가:　 {low}({rate_low})\n거래량: {volume}({rate_volume})'
        format_volumeinfo: Volume information text format. default '{dt}\n\n거래량　　　: {volume}\n거래량증가율: {rate_volume}'

        min_distance: Minimum number of candles that can be selected with the slider. default 30
        limit_candle: Maximum number of candles to draw. default 800
        limit_wick:  Maximum number of candle wicks to draw. default 4,000
        limit_volume: Maximum number of volume bars to draw. default 800. Applies only to drawing candle wicks or price line.
        limit_ma: If the number of displayed data is more than this, the price moving average line is not drawn. default 8,000

        color_navigator_line: Navigator divider color. default '#1e78ff'
        color_navigator_cover: Unselected slider area color. default = 'k'
    """
    pass